from __future__ import annotations

from argrelay.plugin_interp.FuncTreeInterp import FuncTreeInterp
from argrelay.runtime_context.InterpContext import InterpContext


class JumpTreeInterp(FuncTreeInterp):

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        func_ids_to_func_rel_paths: dict[str, list[list[str]]],
        paths_to_jump: dict[tuple[str, ...], tuple[str, ...]],
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
            func_ids_to_func_rel_paths,
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = paths_to_jump

    def select_next_interp_tree_abs_path(self):
        if self.interp_ctx.interp_tree_abs_path in self.paths_to_jump:
            # FS_91_88_07_23 jump tree: replace current `interp_tree_abs_path` with another one based on config:
            self.interp_ctx.interp_tree_abs_path = self.paths_to_jump[self.interp_ctx.interp_tree_abs_path]



from argrelay.plugin_interp.FirstArgInterp import FirstArgInterp
from argrelay.plugin_interp.FirstArgInterpFactoryConfigSchema import (
    first_arg_interp_factory_config_desc,
    first_arg_vals_to_next_interp_factory_ids_,
)
from argrelay.plugin_interp.InterpTreeInterpFactory import InterpTreeInterpFactory
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
    interp_selector_tree_,
)
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig


class FirstArgInterpFactory(InterpTreeInterpFactory):
    """
    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).

    Implements FS_42_76_93_51 first interp.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        first_arg_interp_factory_config_desc.validate_dict(config_dict)
        converted_config = convert_FirstArgInterpConfig_to_InterpTreeInterpFactoryConfig(config_dict)
        super().__init__(
            plugin_instance_id,
            converted_config,
        )

    def activate_plugin(
        self,
        server_config: ServerConfig,
    ):
        """
        # NOTE: FS_42_76_93_51 first interp special case:
        TODO: Think of anther way because `FirstArgInterp` is obsolete - this role here only extends its agony.
        `FirstArgInterp` is used as the root and single instance which
        triggers func loading of itself (and all others recursively).

        Takes part in implementation of FS_01_89_09_24 interp tree.
        """
        interp_tree_abs_path: tuple[str, ...] = tuple([])
        self.load_func_envelopes(
            interp_tree_abs_path,
            server_config,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> FirstArgInterp:
        return FirstArgInterp(
            self.plugin_instance_id,
            # FS_42_76_93_51 first interp special case:
            # We use multiple config clones (with customization) `tree_node_config_dict` for interps at
            # different location on the FS_01_89_09_24 interp tree - these configs get cloned on `load_func_envelopes`.
            # But first interp is always at the root of the interp tree.
            # More over, first interp is the only one who triggers cascading calls to `load_func_envelopes`,
            # but never receives such call itself.
            # Instead of using `interp_tree_node_config_dict`, use `config_dict` directly:
            self.config_dict,
            interp_ctx,
        )


def convert_FirstArgInterpConfig_to_InterpTreeInterpFactoryConfig(config_dict: dict) -> dict:
    """
    `FirstArgInterp` was re-implemented in terms of `InterpTreeInterp` (FS_01_89_09_24).
    """
    interp_selector_tree = {}
    for command_id, plugin_instance_id in config_dict[first_arg_vals_to_next_interp_factory_ids_].items():
        interp_selector_tree[command_id] = plugin_instance_id

    output_dict = {
        interp_selector_tree_: interp_selector_tree
    }

    tree_path_interp_factory_config_desc.validate_dict(output_dict)

    return output_dict



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen = True)
class StaticData:
    """
    All data which normally does not change between requests.

    See also `StaticDataSchema.py`.
    """

    known_arg_types: list = field()
    """
    TODO_26_75_13_27:
    Field `known_arg_types` lists fields of `data_envelop` - they are used to create MongoDB index.
    """

    data_envelopes: list = field()



from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.FuncTreeInterpFactory import FuncTreeInterpFactory
from argrelay.plugin_interp.JumpTreeInterp import JumpTreeInterp
from argrelay.plugin_interp.JumpTreeInterpFactoryConfigSchema import (
    jump_tree_interp_config_desc,
    jump_tree_,
)
from argrelay.plugin_interp.TreeWalker import TreeWalker
from argrelay.runtime_context.InterpContext import InterpContext


class JumpTreeInterpFactory(FuncTreeInterpFactory):
    """
    Implements FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        # FS_91_88_07_23 jump tree
        tree_walker: TreeWalker = TreeWalker(
            "jump",
            self.config_dict[jump_tree_],
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = tree_walker.build_paths_to_paths()
        """
        Implements FS_91_88_07_23 jump tree:
        for given `interp_tree_abs_path` (FS_01_89_09_24) selects next `interp_tree_abs_path`.
        """

    def load_config(
        self,
        config_dict,
    ) -> dict:
        # TODO_74_03_78_60: Call `TypeDesc` API to do load (to populate defaults) -> dump automatically.
        return jump_tree_interp_config_desc.dict_schema.dump(
            jump_tree_interp_config_desc.dict_schema.load(
                config_dict
            )
        )

    def validate_config(
        self,
    ):
        jump_tree_interp_config_desc.validate_dict(self.config_dict)

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        return JumpTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path],
            interp_ctx,
            self.func_ids_to_func_rel_paths,
            self.paths_to_jump,
        )



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class QueryResult:
    """
    Results of a query via `QueryEngine.query_prop_values`.

    Do not confuse with results of `QueryEngine.query_data_envelopes` (which returns all `data_envelope`-s directly).
    """

    data_envelopes: list[dict] = field()
    """
    Contains 0 or 1 `data_envelope`-s from search results.

    *   0: Nothing found.
    *   1: One of the found ones (possibly many).
        If `found_count` > 1, it is the last `data_envelope`-s among many found.
        If `found_count` = 1, it is exactly the `data_envelope` uniquely selected.

    Returning 1 while `found_count` > 1 is used for performance reasons with `QueryEngine.query_prop_values`.

    See also `EnvelopeContainer.data_envelopes`.
    """

    found_count: int = field()
    """
    Total number of `data_envelope`-s found - see `data_envelopes`.
    """

    remaining_types_to_values: dict[str, list[str]] = field()
    """
    See `EnvelopeContainer.remaining_types_to_values`.
    """



from copy import deepcopy

from marshmallow import RAISE, fields

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.plugin_interp.FuncTreeInterpFactoryConfigSchema import (
    FuncTreeInterpFactoryConfigSchema,
    func_tree_interp_config_desc,
)

jump_tree_ = "jump_tree"


class JumpTreeInterpFactoryConfigSchema(FuncTreeInterpFactoryConfigSchema):
    class Meta:
        unknown = RAISE
        strict = True

    # This is a tree (`dict`) of arbitrary depth with `list[str]` leaves.
    # Ideally, this should be defined as nested `dict`,
    # but it is unknown how to do it in marshmallow.
    # Implements FS_91_88_07_23 jump tree.
    jump_tree = fields.Raw(
        required = True,
    )


_re_tree_interp_config_example = deepcopy(func_tree_interp_config_desc.dict_example)
_re_tree_interp_config_example.update({
    jump_tree_: [
        "relay_demo",
    ],
})

jump_tree_interp_config_desc = TypeDesc(
    dict_schema = JumpTreeInterpFactoryConfigSchema(),
    ref_name = JumpTreeInterpFactoryConfigSchema.__name__,
    dict_example = _re_tree_interp_config_example,
    default_file_path = "",
)
